#include <tty/sss_zero.hpp>
#include <fstream>
#include <limits>
#include <cstring>
#include <ctime>
#include <boost/preprocessor.hpp>
#include <boost/random.hpp>
#include <tty/rope_array.hpp>
#include <tty/tasksystem.hpp>
#include <tty/alib.hpp>
#include <tty/autodelete.hpp>
#include <tty/bass.hpp>
#include <tty/cppsdl.hpp>

#define PROGRAMVERSION "v0.01"

bool pollevent(tty::sdltype::event &e){
	return tty::cppsdl::defpollevent(e);
}

struct bullet_pattern_element{
	int frame;
	float omega, spd;
};

typedef tty::alib<float> alib;
namespace sst = tty::tasksystem::sst;
const int field_size = 440, field_pos_x = 105, field_pos_y = 20;
char channel;
bool quitflag = false;
boost::mt19937 rndgen(static_cast<unsigned int>(std::time(0)));
tty::rope_array<bullet_pattern_element> bullet_pattern[2];
tty::bass bass;
tty::cppsdl *cppsdl = reinterpret_cast<tty::cppsdl*>(0);
tty::cppsdl::keys keys;
tty::cppsdl::fps<1000, 60, 30> fps;

#include "pattern.hpp"

struct gametask{
	template<typename T>
	inline static tty::tasksystem::task *taskmain_create_task(){
		return taskmain.create_task<T>();
	}

	template<typename T>
	inline static tty::tasksystem::task *taskmain_delete_task(tty::tasksystem::artask t){
		tty::tasksystem::task *a = t;
		t = t->prev;
		return taskmain.delete_task<T>(a);
	}

	template<typename T>
	inline static tty::tasksystem::task *taskmain_delete_task_pl(tty::tasksystem::task *t){
		return taskmain.delete_task<T>(t);
	}

	template<typename T>
	inline static tty::tasksystem::task *drawmain_create_task(){
		return drawmain.create_task<T>();
	}

	template<typename T>
	inline static tty::tasksystem::task *drawmain_delete_task(tty::tasksystem::artask t){
		tty::tasksystem::task *a = t;
		t = t->prev;
		return drawmain.delete_task<T>(a);
	}

	template<typename T>
	inline static tty::tasksystem::task *drawmain_delete_task_pl(tty::tasksystem::task *t){
		return drawmain.delete_task<T>(t);
	}
#include "taskstruct.hpp"
#include "taskmain.hpp"
};

gametask::taskmain_t gametask::taskmain;
gametask::drawmain_t gametask::drawmain;

int main(int argc, char* argv[]){
	tty::autodelete ad;

	gametask::drawmain_create_task<gametask::outi::draw>();
	gametask::drawmain_create_task<gametask::game::draw>();
	gametask::drawmain_create_task<gametask::spark::draw>();
	gametask::drawmain_create_task<gametask::bullet_b::draw>();
	gametask::drawmain_create_task<gametask::bullet_r::draw>();
	gametask::drawmain_create_task<gametask::myship::draw>();
	gametask::drawmain_create_task<gametask::bullet_echo::draw>();
	gametask::drawmain_create_task<gametask::smile::draw>();
	gametask::drawmain_create_task<gametask::bombed::draw>();
	gametask::drawmain_create_task<gametask::myship::draw_grazecounter>();
	gametask::drawmain_create_task<gametask::blood::draw>();
	gametask::drawmain_create_task<gametask::gameover::draw>();
	gametask::drawmain_create_task<gametask::titledemo::draw>();
	gametask::drawmain_create_task<gametask::titleready::draw>();

	if(argc >= 3){
		bool fullscreen = true;
		for(int i = 3; i < argc; ++i){
			if(std::strcmp(argv[i], "-window") == 0){
				fullscreen = false;
			}else if(std::strcmp(argv[i], "-fullscreen") == 0){
				fullscreen = true;
			}
		}
		cppsdl = new tty::cppsdl("otodama " PROGRAMVERSION " - test_void", 640, 480, 32, 0, fullscreen ? tty::sdlval::sdl::fullscreen : 0, pollevent);
		if(fullscreen){
			cppsdl->warpmouse(40, 375);
		}
		if(!bass.streamcreatefile(argv[1])){
			return 0;
		}
#define READ(a) f.read(reinterpret_cast<char*>(&a), sizeof(a)); if(f.fail()){ return 0; }
		std::fstream f(argv[2], std::ios::in | std::ios::binary);
		if(f.fail()){
			return 0;
		}
		bullet_pattern_element t;
		READ(channel);
		if((channel != 1) && (channel != 2)){
			return 0;
		}
		for(int cc = 0; cc < channel; ++cc){
			for(; ; ){
				READ(t.frame);
				if(t.frame == -1){
					break;
				}
				READ(t.omega);
				READ(t.spd);
				bullet_pattern[cc].push_back(t);
			}
		}
#undef READ
		gametask::taskmain_create_task<gametask::titleready>();
	}else{
		cppsdl = new tty::cppsdl("otodama " PROGRAMVERSION " - test_void", 640, 480, 32, 0, tty::sdlval::sdl::fullscreen, pollevent);
		cppsdl->warpmouse(40, 375);
		gametask::taskmain_create_task<gametask::titledemo>();
	}

	ad.set(cppsdl);

	if(!pattern::p.test()){
		return 0;
	}

	pattern::p.setcolorkey(
		tty::sdlval::sdl::srccolorkey | tty::sdlval::sdl::rleaccel,
		cppsdl->maprgb(0x00, 0xFF, 0xFF)
	);

	for(; ; ){
		keys.update();
		if(!cppsdl->fn() || quitflag){
			break;
		}
		cppsdl->fillrect(0, 0, cppsdl->screen_width(), cppsdl->screen_height(), 0x000000);
		cppsdl->fillrect(105, 20, 440, 440, 0xFFFFFF);
		cppsdl->fillrect(106, 21, 438, 438, 0x000000);
		gametask::taskmain();
		gametask::drawmain();
		cppsdl->flip();
		fps();
	}

	return 0;
}
